/* arrow.c: ascii2xfig arrow parser and renderer
 *
 * Copyright (C) 2006 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

//#define KDEBUG
#include "defs.h"

#define ___ 1	/* same line */

#define UP_ 0	/* up a row */
#define DWN 2	/* down a row */

#define LFT 0	/* left a column */
#define RGT 2	/* right a column */

static void extract_arrow(struct grid *grid, int row, int col);
static void find_arrow_end(struct grid *grid, struct arrow *arrow);
static void follow_arrow_left(struct grid *grid, struct arrow *arrow, struct point *last);
static void follow_arrow_right(struct grid *grid, struct arrow *arrow, struct point *last);
static void follow_arrow_up(struct grid *grid, struct arrow *arrow, struct point *last);
static void follow_arrow_down(struct grid *grid, struct arrow *arrow, struct point *last);
static void follow_arrow_up_left(struct grid *grid, struct arrow *arrow, struct point *last);
static void follow_arrow_up_right(struct grid *grid, struct arrow *arrow, struct point *last);
static void follow_arrow_down_left(struct grid *grid, struct arrow *arrow, struct point *last);
static void follow_arrow_down_right(struct grid *grid, struct arrow *arrow, struct point *last);
static void follow_arrow_junction(struct grid *grid, struct arrow *arrow, struct point *last);
static void shorten_arrow_diagonals(struct grid *grid, struct arrow *arrow);
static void free_arrow_point(struct point *point);
static void free_arrow(struct arrow *arrow);

static const char *ends[] = {
	[A_LEFT ]	= "left",
	[A_RIGHT]	= "right",
	[A_UP   ]	= "up",
	[A_DOWN ]	= "down",
	[A_UP_LEFT]	= "up-left",
	[A_UP_RIGHT]	= "up-right",
	[A_DOWN_LEFT]	= "down-left",
	[A_DOWN_RIGHT]	= "down-right",
	[A_OTHER]	= "other",
};

/*****************************************************************************/
/*
 * find all the arrows in the diagram
 */
void find_arrows(struct grid *grid)
{
	char **lines = grid->lines;
	int row, col;

	/* search for the left/top end of an arrow */
	for (row = 0; row < grid->nlines; row++) {
		for (col = 0; col < grid->ncols; col++) {
			if (grid->markup[row][col])
				continue;

			if (strchr("/|\\-<>^V", lines[row][col]))
				extract_arrow(grid, row, col);
		}
	}
}

/*****************************************************************************/
/*
 * attempt to extract an arrow from the diagram at a given row and column
 * - an arrow starts and ends at one of "/|\-<>^V"
 * - an arrow may transit through one of "/|\-"
 * - an arrow may skip, eg: "---|---"
 * - an arrow may have a junction or a corner at a "+"
 * - an arrow terminates at a box boundary, a space, text or an arrowhead
 */
static void extract_arrow(struct grid *grid, int row, int col)
{
	struct arrow *arrow;
	char **lines = grid->lines;

	arrow = xmalloc(sizeof(*arrow));
	arrow->row = row;
	arrow->col = col;
	arrow->end = A_OTHER;
	arrow->shead = 0;
	arrow->ehead = 0;
	arrow->npoints = 1;
	arrow->start.next = NULL;

	_debug("\e[44mFOLLOW___________________________________________\e[m");
	find_arrow_end(grid, arrow);

	_debug("\e[42mARROW: %2d,%2d: '%c' %s\e[m",
	       row, col, lines[arrow->row][arrow->col], ends[arrow->end]);

	/* start by moving away from the arrow end we know we've got */
	switch (arrow->end) {
	case A_LEFT:
		arrow->start.x = XSCALEH(arrow->col - 1);
		arrow->start.y = YSCALEH(arrow->row);
		follow_arrow_right(grid, arrow, &arrow->start);
		break;
	case A_RIGHT:
		arrow->start.x = XSCALEH(arrow->col + 1);
		arrow->start.y = YSCALEH(arrow->row);
		follow_arrow_left(grid, arrow, &arrow->start);
		break;
	case A_UP:
		arrow->start.x = XSCALEH(arrow->col);
		arrow->start.y = YSCALEH(arrow->row - 1);
		if (grid->markup[arrow->row - 1][arrow->col] & MARK_TEXT)
			arrow->start.y = YSCALEHF(arrow->row, 4, 10);
		follow_arrow_down(grid, arrow, &arrow->start);
		break;
	case A_DOWN:
		arrow->start.x = XSCALEH(arrow->col);
		arrow->start.y = YSCALEH(arrow->row + 1);
		if (grid->markup[arrow->row + 1][arrow->col] & MARK_TEXT)
			arrow->start.y = YSCALEHF(arrow->row, 4, 10);
		follow_arrow_up(grid, arrow, &arrow->start);
		break;
	case A_UP_LEFT:
		arrow->start.x = XSCALEH(arrow->col - 1);
		arrow->start.y = YSCALEH(arrow->row - 1);
		follow_arrow_down_right(grid, arrow, &arrow->start);
		break;
	case A_UP_RIGHT:
		arrow->start.x = XSCALEH(arrow->col + 1);
		arrow->start.y = YSCALEH(arrow->row - 1);
		follow_arrow_up_left(grid, arrow, &arrow->start);
		break;
	case A_DOWN_LEFT:
		arrow->start.x = XSCALEH(arrow->col - 1);
		arrow->start.y = YSCALEH(arrow->row + 1);
		follow_arrow_up_right(grid, arrow, &arrow->start);
		break;
	case A_DOWN_RIGHT:
		arrow->start.x = XSCALEH(arrow->col + 1);
		arrow->start.y = YSCALEH(arrow->row + 1);
		follow_arrow_up_left(grid, arrow, &arrow->start);
		break;
	default:
		return;
	}

	shorten_arrow_diagonals(grid, arrow);

	arrow->link = grid->arrows;
	grid->arrows = arrow;
}

/*****************************************************************************/
/*
 * map the arrow steps into a convenient 3x3 grid
 */
static void map_arrow_steps(struct grid *grid, struct arrow *arrow, int row, int col)
{
	unsigned char **markup = grid->markup;
	char **lines = grid->lines;

	arrow->steps[UP_][LFT] = lines[row - 1][col - 1];
	arrow->steps[UP_][___] = lines[row - 1][col    ];
	arrow->steps[UP_][RGT] = lines[row - 1][col + 1];
	arrow->steps[___][LFT] = lines[row    ][col - 1];
	arrow->steps[___][___] = lines[row    ][col    ];
	arrow->steps[___][RGT] = lines[row    ][col + 1];
	arrow->steps[DWN][LFT] = lines[row + 1][col - 1];
	arrow->steps[DWN][___] = lines[row + 1][col    ];
	arrow->steps[DWN][RGT] = lines[row + 1][col + 1];

	if (markup[row - 1][col - 1]) arrow->steps[UP_][LFT] = 0;
	if (markup[row - 1][col    ]) arrow->steps[UP_][___] = 0;
	if (markup[row - 1][col + 1]) arrow->steps[UP_][RGT] = 0;
	if (markup[row    ][col - 1]) arrow->steps[___][LFT] = 0;
	if (markup[row    ][col    ]) arrow->steps[___][___] = 0;
	if (markup[row    ][col + 1]) arrow->steps[___][RGT] = 0;
	if (markup[row + 1][col - 1]) arrow->steps[DWN][LFT] = 0;
	if (markup[row + 1][col    ]) arrow->steps[DWN][___] = 0;
	if (markup[row + 1][col + 1]) arrow->steps[DWN][RGT] = 0;

	arrow->psteps[ 0] = arrow->steps[UP_][LFT] ?: ' ';
	arrow->psteps[ 1] = arrow->steps[UP_][___] ?: ' ';
	arrow->psteps[ 2] = arrow->steps[UP_][RGT] ?: ' ';
	arrow->psteps[ 3] = '.';
	arrow->psteps[ 4] = arrow->steps[___][LFT] ?: ' ';
	arrow->psteps[ 5] = arrow->steps[___][___] ?: ' ';
	arrow->psteps[ 6] = arrow->steps[___][RGT] ?: ' ';
	arrow->psteps[ 7] = '.';
	arrow->psteps[ 8] = arrow->steps[DWN][LFT] ?: ' ';
	arrow->psteps[ 9] = arrow->steps[DWN][___] ?: ' ';
	arrow->psteps[10] = arrow->steps[DWN][RGT] ?: ' ';
	arrow->psteps[11] = '\0';
}

/*****************************************************************************/
/*
 * find one end of the arrow so that we don't have to handle it from the middle
 */
static void find_arrow_end(struct grid *grid, struct arrow *arrow)
{
	arrow_t at = A_OTHER;
	char **lines = grid->lines;
	int row = arrow->row, col = arrow->col, dr = 0, dc = 0;

	map_arrow_steps(grid, arrow, row, col);
	_debug("- find: %2d,%2d: '%c' [%s]",
	       row, col, lines[row][col], arrow->psteps);

	switch (lines[row][col]) {
		/* arrowheads */
	case '<': arrow->end = A_LEFT;	goto found_end;
	case '>': arrow->end = A_RIGHT;	goto found_end;
	case '^': arrow->end = A_UP;	goto found_end;
	case 'V': arrow->end = A_DOWN;	goto found_end;

		/* horizontal line */
	case '-':
		switch (arrow->steps[___][LFT]) {
		case '<': goto transit_left;
		case '-': goto transit_left;
		case '+': goto transit_left;
		}
		switch (arrow->steps[___][RGT]) {
		case '>': goto transit_right;
		case '-': goto transit_right;
		case '+': goto transit_right;
		}

		if (arrow->steps[UP_][LFT] == '\\')	goto transit_up_left;
		if (arrow->steps[UP_][RGT] == '/')	goto transit_up_right;
		if (arrow->steps[DWN][LFT] == '/')	goto transit_down_left;
		if (arrow->steps[DWN][RGT] == '\\')	goto transit_down_right;

		switch (lines[row][col - 1]) {
		case '|':
		case '/':
		case '\\':
			/* is this a skip */
			_debug("- skipLL? '%c' %02x, '%c' %02x",
			       lines[row][col - 1],
			       grid->markup[row][col - 1],
			       lines[row][col - 2],
			       grid->markup[row][col - 2]);

			if (!(grid->markup[row][col - 1] & ~MARK_ARROW) &&
			    !(grid->markup[row][col - 2] & ~MARK_SEEN) &&
			    lines[row][col - 2] == '-'
			    ) {
				if (grid->markup[row][col - 2] & MARK_SEEN)
					break; /* just come from there */
				_debug("- skipping");
				goto transit_skip_left;
			}
			goto found_end;
		}

		switch (lines[row][col + 1]) {
		case '|':
		case '/':
		case '\\':
			/* is this a skip */
			_debug("- skipRR? '%c' %02x, '%c' %02x",
			       lines[row][col + 1],
			       grid->markup[row][col + 1],
			       lines[row][col + 2],
			       grid->markup[row][col + 2]);

			if (!(grid->markup[row][col + 1] & ~MARK_ARROW) &&
			    !(grid->markup[row][col + 2] & ~MARK_SEEN) &&
			    lines[row][col + 2] == '-'
			    ) {
				if (grid->markup[row][col + 2] & MARK_SEEN)
					break; /* just come from there */
				_debug("- skipping");
				goto transit_skip_right;
			}
			goto found_end;
		}

		goto found_end;

		/* vertical line */
	case '|':
		switch (arrow->steps[UP_][___]) {
		case '^': goto transit_up;
		case '|': goto transit_up;
		case '+': goto transit_up;
		}
		switch (arrow->steps[DWN][___]) {
		case 'V': goto transit_down;
		case '|': goto transit_down;
		case '+': goto transit_down;
		}

		if (arrow->steps[UP_][LFT] == '\\')	goto transit_up_left;
		if (arrow->steps[UP_][RGT] == '/')	goto transit_up_right;
		if (arrow->steps[DWN][LFT] == '/')	goto transit_down_left;
		if (arrow->steps[DWN][RGT] == '\\')	goto transit_down_right;

		switch (lines[row - 1][col]) {
		case '-':
		case '/':
		case '\\':
			/* is this a skip */
			if (!(grid->markup[row - 1][col] & ~MARK_ARROW) &&
			    !(grid->markup[row - 2][col] & ~MARK_SEEN) &&
			    lines[row - 2][col] == '|'
			    ) {
				if (grid->markup[row - 2][col] & MARK_SEEN)
					break; /* just come from there */
				goto transit_skip_up;
			}
			goto found_end;
		}
		switch (lines[row + 1][col]) {
		case '-':
		case '/':
		case '\\':
			/* is this a skip */
			if (!(grid->markup[row + 1][col] & ~MARK_ARROW) &&
			    !(grid->markup[row + 2][col] & ~MARK_SEEN) &&
			    lines[row + 2][col] == '|'
			    ) {
				if (grid->markup[row + 2][col] & MARK_SEEN)
					break; /* just come from there */
				goto transit_skip_down;
			}
			goto found_end;
		}

		goto found_end;

		/* lowleft to highright diagonal line */
	case '/':
		switch (arrow->steps[UP_][RGT]) {
		case '/': goto transit_up_right;
		case '|': goto transit_up_right;
		case '-': goto transit_up_right;
		}
		switch (arrow->steps[DWN][LFT]) {
		case '/': goto transit_down_left;
		case '|': goto transit_down_left;
		case '-': goto transit_down_left;
		}

		switch (lines[row - 1][col + 1]) {
		case '\\':
			/* is this a skip */
			if (!(grid->markup[row - 1][col + 1] & ~MARK_ARROW) &&
			    !(grid->markup[row - 2][col + 2] & ~MARK_SEEN) &&
			    lines[row - 2][col + 2] == '/'
			    ) {
				if (grid->markup[row - 2][col + 2] & MARK_SEEN)
					break; /* just come from there */
				goto transit_skip_up_right;
			}
			goto found_end;
		}
		switch (lines[row + 1][col - 1]) {
		case '\\':
			/* is this a skip */
			if (!(grid->markup[row + 1][col - 1] & ~MARK_ARROW) &&
			    !(grid->markup[row + 2][col - 2] & ~MARK_SEEN) &&
			    lines[row + 2][col - 2] == '/'
			    ) {
				if (grid->markup[row + 2][col - 2] & MARK_SEEN)
					break; /* just come from there */
				goto transit_skip_down_left;
			}
			goto found_end;
		}

		if (arrow->steps[___][LFT] == '\\')
			goto transit_left;
		if (arrow->steps[___][RGT] == '\\')
			goto transit_right;
		if (arrow->steps[UP_][___] == '\\')
			goto transit_up;
		if (arrow->steps[DWN][___] == '\\')
			goto transit_down;

		goto found_end;

		/* highleft to lowright diagonal line */
	case '\\':
		switch (arrow->steps[UP_][LFT]) {
		case '\\': goto transit_up_left;
		case '|': goto transit_up_left;
		case '-': goto transit_up_left;
		}
		switch (arrow->steps[DWN][RGT]) {
		case '\\': goto transit_down_right;
		case '|': goto transit_down_right;
		case '-': goto transit_down_right;
		}

		if (arrow->steps[UP_][LFT] == '\\')	goto transit_up_left;
		if (arrow->steps[UP_][RGT] == '/')	goto transit_up_right;
		if (arrow->steps[DWN][LFT] == '/')	goto transit_down_left;
		if (arrow->steps[DWN][RGT] == '\\')	goto transit_down_right;

		switch (lines[row - 1][col - 1]) {
		case '/':
			/* is this a skip */
			if (!(grid->markup[row - 1][col - 1] & ~MARK_ARROW) &&
			    !(grid->markup[row - 2][col - 2] & ~MARK_SEEN) &&
			    lines[row - 2][col - 2] == '\\'
			    ) {
				if (grid->markup[row - 2][col - 2] & MARK_SEEN)
					break; /* just come from there */
				goto transit_skip_up_left;
			}
			goto found_end;
		}
		switch (lines[row + 1][col + 1]) {
		case '/':
			/* is this a skip */
			if (!(grid->markup[row + 1][col + 1] & ~MARK_ARROW) &&
			    !(grid->markup[row + 2][col + 2] & ~MARK_SEEN) &&
			    lines[row + 2][col + 2] == '\\'
			    ) {
				if (grid->markup[row + 2][col + 2] & MARK_SEEN)
					break; /* just come from there */
				goto transit_skip_down_right;
			}
			goto found_end;
		}

		if (arrow->steps[___][LFT] == '/')
			goto transit_left;
		if (arrow->steps[___][RGT] == '/')
			goto transit_right;
		if (arrow->steps[UP_][___] == '/')
			goto transit_up;
		if (arrow->steps[DWN][___] == '/')
			goto transit_down;

		goto found_end;

		/* junction */
	case '+':
		if (arrow->steps[UP_][___] == '|') goto transit_up;
		if (arrow->steps[DWN][___] == '|') goto transit_down;
		if (arrow->steps[___][LFT] == '-') goto transit_left;
		if (arrow->steps[___][RGT] == '-') goto transit_right;
		goto found_end;

	default:
		goto found_end;
	}

	goto found_end;

transit_up_left:		dr = -1;	dc = -1;	at = A_UP_LEFT;	goto transit;
transit_up_right:		dr = -1;	dc = +1;	at = A_UP_RIGHT; goto transit;
transit_left:					dc = -1;	at = A_LEFT;	goto transit;
transit_right:					dc = +1;	at = A_RIGHT;	goto transit;
transit_down_left:		dr = +1;	dc = -1;	at = A_DOWN_LEFT; goto transit;
transit_down_right:		dr = +1;	dc = +1;	at = A_DOWN_RIGHT; goto transit;
transit_up:			dr = -1;			at = A_UP;	goto transit;
transit_down:			dr = +1;			at = A_DOWN;	goto transit;

transit_skip_up_left:		dr = -2;	dc = -2;	at = A_UP_LEFT;	goto transit;
transit_skip_up_right:		dr = -2;	dc = +2;	at = A_UP_RIGHT; goto transit;
transit_skip_left:				dc = -2;	at = A_LEFT;	goto transit;
transit_skip_right:				dc = +2;	at = A_RIGHT;	goto transit;
transit_skip_down_left:		dr = +2;	dc = -2;	at = A_DOWN_LEFT; goto transit;
transit_skip_down_right:	dr = +2;	dc = +2;	at = A_DOWN_RIGHT; goto transit;
transit_skip_up:		dr = -2;			at = A_UP;	goto transit;
transit_skip_down:		dr = +2;			at = A_DOWN;	goto transit;

transit:
	/* recurse to next part of arrow
	 * - leave a temporary mark to make sure we don't loop
	 */
	_debug("- transit %d,%d: '%c'",
	       dr, dc, grid->lines[row + dr][col + dc]);

	assert(strchr("/|\\-+<>V^", grid->lines[row + dr][col + dc]));

	grid->markup[row][col] |= MARK_SEEN;
	arrow->row += dr;
	arrow->col += dc;
	arrow->end = at;
	find_arrow_end(grid, arrow);
	grid->markup[row][col] &= ~MARK_SEEN;

found_end:
	return;
}

/*****************************************************************************/
/*
 * follow an arrow leftwards from right-hand point
 */
static void follow_arrow_left(struct grid *grid, struct arrow *arrow, struct point *last)
{
	struct point *pt;
	int row = arrow->row, col = arrow->col;

	pt = xmalloc(sizeof(*pt));
	pt->skip = 0;
	pt->branch = NULL;
	pt->next = NULL;
	last->next = pt;
	arrow->npoints++;

	if (grid->lines[row][col] == '>')
		arrow->shead = 1;

resume:
	/* follow left as far as possible */
	for (;;) {
		map_arrow_steps(grid, arrow, row, col);
		_debug("- LL step: %2d,%2d: '%c' [%s]",
		       row, col, arrow->steps[___][___], arrow->psteps);
		grid->markup[row][col] |= MARK_ARROW;
		if (arrow->steps[___][LFT] != '-')
			break;
		col--;
	}

	arrow->row = row;
	arrow->col = col - 1;

	/* the point is in the middle of the next box left */
	pt->x = XSCALEH(col - 1);
	pt->y = last->y;

	/* stop if we've found the other head */
	if (arrow->steps[___][LFT] == '<') {
		grid->markup[row][col - 1] |= MARK_ARROW;
		arrow->ehead = 1;
		pt->x -= XSCALE(1);
		return;
	}

	/* handle a junction or a corner */
	if (arrow->steps[___][LFT] == '+') {
		follow_arrow_junction(grid, arrow, pt);
		return;
	}

	/* handle left up/down diagonals */
	if (arrow->steps[UP_][LFT] == '\\') {
		arrow->row--;
		pt->x += XSCALE(1);
		follow_arrow_up_left(grid, arrow, pt);
		return;
	}

	if (arrow->steps[DWN][LFT] == '/') {
		arrow->row++;
		pt->x += XSCALE(1);
		follow_arrow_down_left(grid, arrow, pt);
		return;
	}

	/* handle a skip */
	if (col > 2 &&
	    grid->lines[row][col - 1] == '|' &&
	    grid->lines[row][col - 2] == '-'
	    ) {
		col -= 2;
		goto resume;
	}
}

/*****************************************************************************/
/*
 * follow an arrow rightwards
 */
static void follow_arrow_right(struct grid *grid, struct arrow *arrow, struct point *last)
{
	struct point *pt;
	int row = arrow->row, col = arrow->col;

	pt = xmalloc(sizeof(*pt));
	pt->skip = 0;
	pt->branch = NULL;
	pt->next = NULL;
	last->next = pt;
	arrow->npoints++;

	if (grid->lines[row][col] == '<')
		arrow->shead = 1;

resume:
	/* follow right as far as possible */
	for (;;) {
		map_arrow_steps(grid, arrow, row, col);
		_debug("- RR step: %2d,%2d: '%c' [%s]",
		       row, col, arrow->steps[___][___], arrow->psteps);
		grid->markup[row][col] |= MARK_ARROW;
		if (arrow->steps[___][RGT] != '-')
			break;
		col++;
	}

	arrow->row = row;
	arrow->col = col + 1;

	/* the point is in the middle of the next box right */
	pt->x = XSCALEH(col + 1);
	pt->y = last->y;

	/* stop if we've found the other head */
	if (arrow->steps[___][RGT] == '>') {
		grid->markup[row][col + 1] |= MARK_ARROW;
		arrow->ehead = 1;
		pt->x += XSCALE(1);
		return;
	}

	/* handle a junction or a corner */
	if (arrow->steps[___][RGT] == '+') {
		follow_arrow_junction(grid, arrow, pt);
		return;
	}

	/* handle left up/down diagonals */
	if (arrow->steps[UP_][RGT] == '/') {
		arrow->row--;
		pt->x -= XSCALE(1);
		follow_arrow_up_right(grid, arrow, pt);
		return;
	}

	if (arrow->steps[DWN][RGT] == '\\') {
		arrow->row++;
		pt->x -= XSCALE(1);
		follow_arrow_down_right(grid, arrow, pt);
		return;
	}

	/* handle a skip */
	if (col < grid->ncols - 2 &&
	    grid->lines[row][col + 1] == '|' &&
	    grid->lines[row][col + 2] == '-'
	    ) {
		col += 2;
		goto resume;
	}
}

/*****************************************************************************/
/*
 * follow an arrow upwards
 */
static void follow_arrow_up(struct grid *grid, struct arrow *arrow, struct point *last)
{
	struct point *pt;
	int row = arrow->row, col = arrow->col;

	pt = xmalloc(sizeof(*pt));
	pt->skip = 0;
	pt->branch = NULL;
	pt->next = NULL;
	last->next = pt;
	arrow->npoints++;

	if (grid->lines[row][col] == 'V')
		arrow->shead = 1;

resume:
	/* follow up as far as possible */
	for (;;) {
		map_arrow_steps(grid, arrow, row, col);
		_debug("- UU step: %2d,%2d: '%c' [%s]",
		       row, col, arrow->steps[___][___], arrow->psteps);
		grid->markup[row][col] |= MARK_ARROW;
		if (arrow->steps[UP_][___] != '|')
			break;
		row--;
	}

	arrow->row = row - 1;
	arrow->col = col;

	/* the point is in the middle of the next box up */
	pt->x = last->x;
	pt->y = YSCALEH(row - 1);

	/* stop if we've found the other head */
	if (arrow->steps[UP_][___] == '^') {
		grid->markup[row - 1][col] |= MARK_ARROW;
		arrow->ehead = 1;
		pt->y -= YSCALE(1);
		return;
	}

	/* handle a junction or a corner */
	if (arrow->steps[UP_][___] == '+') {
		follow_arrow_junction(grid, arrow, pt);
		return;
	}

	/* handle left/right up diagonals */
	if (arrow->steps[UP_][LFT] == '\\') {
		arrow->col--;
		pt->y += YSCALE(1);
		follow_arrow_up_left(grid, arrow, pt);
		return;
	}

	if (arrow->steps[UP_][RGT] == '/') {
		arrow->col++;
		pt->y += YSCALE(1);
		follow_arrow_up_right(grid, arrow, pt);
		return;
	}

	/* don't overrun onto text labels */
	if (grid->markup[arrow->row][arrow->col] & MARK_TEXT) {
		pt->x = last->x;
		pt->y = YSCALEHF(row, -4, 10);
	}

	/* handle a skip */
	if (row > 2 &&
	    grid->lines[row - 1][col] == '-' &&
	    grid->lines[row - 2][col] == '|'
	    ) {
		row -= 2;
		goto resume;
	}
}

/*****************************************************************************/
/*
 * follow an arrow downwards
 */
static void follow_arrow_down(struct grid *grid, struct arrow *arrow, struct point *last)
{
	struct point *pt;
	int row = arrow->row, col = arrow->col;

	pt = xmalloc(sizeof(*pt));
	pt->skip = 0;
	pt->branch = NULL;
	pt->next = NULL;
	last->next = pt;
	arrow->npoints++;

	if (grid->lines[row][col] == '^')
		arrow->shead = 1;

resume:
	/* follow down as far as possible */
	for (;;) {
		map_arrow_steps(grid, arrow, row, col);
		_debug("- DD step: %2d,%2d: '%c' [%s]",
		       row, col, arrow->steps[___][___], arrow->psteps);
		grid->markup[row][col] |= MARK_ARROW;
		if (arrow->steps[DWN][___] != '|')
			break;
		row++;
	}

	arrow->row = row + 1;
	arrow->col = col;

	/* the point is in the middle of the next box down */
	pt->x = last->x;
	pt->y = YSCALEH(row + 1);

	/* stop if we've found the other head */
	if (arrow->steps[DWN][___] == 'V') {
		grid->markup[row + 1][col] |= MARK_ARROW;
		arrow->ehead = 1;
		pt->y += YSCALE(1);
		return;
	}

	/* handle a junction or a corner */
	if (arrow->steps[DWN][___] == '+') {
		follow_arrow_junction(grid, arrow, pt);
		return;
	}

	/* handle left/right down diagonals */
	if (arrow->steps[DWN][LFT] == '/') {
		arrow->col--;
		pt->y -= YSCALE(1);
		follow_arrow_down_left(grid, arrow, pt);
		return;
	}

	if (arrow->steps[DWN][RGT] == '\\') {
		arrow->col++;
		pt->y -= YSCALE(1);
		follow_arrow_down_right(grid, arrow, pt);
		return;
	}
	/* don't overrun onto text labels */
	if (grid->markup[arrow->row][arrow->col] & MARK_TEXT) {
		pt->x = last->x;
		pt->y = YSCALEHF(row, 4, 10);
	}

	/* handle a skip */
	if (row < grid->nlines - 2 &&
	    grid->lines[row + 1][col] == '-' &&
	    grid->lines[row + 2][col] == '|'
	    ) {
		row += 2;
		goto resume;
	}
}

/*****************************************************************************/
/*
 * follow an arrow point diagonally up and to the left
 */
static void follow_arrow_up_left(struct grid *grid, struct arrow *arrow, struct point *last)
{
	struct point *pt;
	int row = arrow->row, col = arrow->col;

	pt = xmalloc(sizeof(*pt));
	pt->skip = 0;
	pt->branch = NULL;
	pt->next = NULL;
	last->next = pt;
	arrow->npoints++;

resume:
	/* follow up-left as far as possible */
	for (;;) {
		map_arrow_steps(grid, arrow, row, col);
		_debug("- UL step: %2d,%2d: '%c' [%s]",
		       row, col, arrow->steps[___][___], arrow->psteps);
		grid->markup[row][col] |= MARK_ARROW;
		if (arrow->steps[UP_][LFT] != '\\')
			break;
		row--;
		col--;
	}

	arrow->row = row - 1;
	arrow->col = col - 1;

	/* the point is in the middle of the next box up and left */
	pt->x = XSCALEH(col - 1);
	pt->y = YSCALEH(row - 1);

	/* handle going horizontal or vertical */
	if (arrow->steps[UP_][LFT] == '|') {
		follow_arrow_up(grid, arrow, pt);
		return;
	}

	if (arrow->steps[UP_][LFT] == '-') {
		follow_arrow_left(grid, arrow, pt);
		return;
	}

	/* handle going down-left */
	if (arrow->steps[___][LFT] == '/') {
		arrow->row = row;
		pt->x += XHSCALE(1);
		pt->y += YHSCALE(1);
		follow_arrow_down_left(grid, arrow, pt);
		return;
	}

	/* handle going up-right */
	if (arrow->steps[UP_][___] == '/') {
		arrow->row = row;
		pt->x += XHSCALE(1);
		pt->y += YHSCALE(1);
		follow_arrow_up_right(grid, arrow, pt);
		return;
	}

	/* handle a skip */
	if (row < grid->nlines - 2 &&
	    col < grid->ncols - 2 &&
	    grid->lines[row + 1][col + 1] == '|' &&
	    grid->lines[row + 2][col + 2] == '\\'
	    ) {
		row += 2;
		col += 2;
		goto resume;
	}
}

/*****************************************************************************/
/*
 * follow an arrow point diagonally up and to the right
 */
static void follow_arrow_up_right(struct grid *grid, struct arrow *arrow, struct point *last)
{
	struct point *pt;
	int row = arrow->row, col = arrow->col;

	pt = xmalloc(sizeof(*pt));
	pt->skip = 0;
	pt->branch = NULL;
	pt->next = NULL;
	last->next = pt;
	arrow->npoints++;

resume:
	/* follow up-right as far as possible */
	for (;;) {
		map_arrow_steps(grid, arrow, row, col);
		_debug("- UR step: %2d,%2d: '%c' [%s]",
		       row, col, arrow->steps[___][___], arrow->psteps);
		grid->markup[row][col] |= MARK_ARROW;
		if (arrow->steps[UP_][RGT] != '/')
			break;
		row--;
		col++;
	}

	arrow->row = row - 1;
	arrow->col = col + 1;

	/* the point is in the middle of the next box up and right */
	pt->x = XSCALEH(col + 1);
	pt->y = YSCALEH(row - 1);

	/* handle going horizontal or vertical */
	if (arrow->steps[UP_][RGT] == '|') {
		follow_arrow_up(grid, arrow, pt);
		return;
	}

	if (arrow->steps[UP_][RGT] == '-') {
		follow_arrow_right(grid, arrow, pt);
		return;
	}

	/* handle going down-right */
	if (arrow->steps[___][RGT] == '\\') {
		arrow->row = row;
		pt->x -= XHSCALE(1);
		pt->y += YHSCALE(1);
		follow_arrow_down_right(grid, arrow, pt);
		return;
	}

	/* handle going up-left */
	if (arrow->steps[UP_][___] == '\\') {
		arrow->row = row;
		pt->x -= XHSCALE(1);
		pt->y += YHSCALE(1);
		follow_arrow_up_left(grid, arrow, pt);
		return;
	}

	/* handle a skip */
	if (row > 2 &&
	    col < grid->ncols - 2 &&
	    grid->lines[row - 1][col + 1] == '|' &&
	    grid->lines[row - 2][col + 2] == '/'
	    ) {
		row -= 2;
		col += 2;
		goto resume;
	}
}

/*****************************************************************************/
/*
 * follow an arrow point diagonally down and to the left
 */
static void follow_arrow_down_left(struct grid *grid, struct arrow *arrow, struct point *last)
{
	struct point *pt;
	int row = arrow->row, col = arrow->col;

	pt = xmalloc(sizeof(*pt));
	pt->skip = 0;
	pt->branch = NULL;
	pt->next = NULL;
	last->next = pt;
	arrow->npoints++;

resume:
	/* follow down-left as far as possible */
	for (;;) {
		map_arrow_steps(grid, arrow, row, col);
		_debug("- DL step: %2d,%2d: '%c' [%s]",
		       row, col, arrow->steps[___][___], arrow->psteps);
		grid->markup[row][col] |= MARK_ARROW;
		if (arrow->steps[DWN][LFT] != '/')
			break;
		row++;
		col--;
	}

	arrow->row = row + 1;
	arrow->col = col - 1;

	/* the point is in the middle of the next box up and left */
	pt->x = XSCALEH(col - 1);
	pt->y = YSCALEH(row + 1);

	/* handle going horizontal or vertical */
	if (arrow->steps[DWN][LFT] == '|') {
		follow_arrow_down(grid, arrow, pt);
		return;
	}

	if (arrow->steps[DWN][LFT] == '-') {
		follow_arrow_left(grid, arrow, pt);
		return;
	}

	/* handle going up-left */
	if (arrow->steps[___][LFT] == '\\') {
		arrow->row = row;
		pt->x += XHSCALE(1);
		pt->y -= YHSCALE(1);
		follow_arrow_up_left(grid, arrow, pt);
		return;
	}

	/* handle going down-right */
	if (arrow->steps[RGT][___] == '\\') {
		arrow->row = row;
		pt->x += XHSCALE(1);
		pt->y -= YHSCALE(1);
		follow_arrow_down_right(grid, arrow, pt);
		return;
	}

	/* handle a skip */
	if (row < grid->nlines - 2 &&
	    col < grid->ncols - 2 &&
	    grid->lines[row + 1][col + 1] == '|' &&
	    grid->lines[row + 2][col + 2] == '/'
	    ) {
		row += 2;
		col += 2;
		goto resume;
	}
}

/*****************************************************************************/
/*
 * follow an arrow point diagonally down and to the right
 */
static void follow_arrow_down_right(struct grid *grid, struct arrow *arrow, struct point *last)
{
	struct point *pt;
	int row = arrow->row, col = arrow->col;

	pt = xmalloc(sizeof(*pt));
	pt->skip = 0;
	pt->branch = NULL;
	pt->next = NULL;
	last->next = pt;
	arrow->npoints++;

resume:
	/* follow down-right as far as possible */
	for (;;) {
		map_arrow_steps(grid, arrow, row, col);
		_debug("- DR step: %2d,%2d: '%c' [%s]",
		       row, col, arrow->steps[___][___], arrow->psteps);
		grid->markup[row][col] |= MARK_ARROW;
		if (arrow->steps[DWN][RGT] != '\\')
			break;
		row++;
		col++;
	}

	arrow->row = row + 1;
	arrow->col = col + 1;

	/* the point is in the middle of the next box up and right */
	pt->x = XSCALEH(col + 1);
	pt->y = YSCALEH(row + 1);

	/* handle going horizontal or vertical */
	if (arrow->steps[DWN][RGT] == '|') {
		follow_arrow_down(grid, arrow, pt);
		return;
	}

	if (arrow->steps[DWN][RGT] == '-') {
		follow_arrow_right(grid, arrow, pt);
		return;
	}

	/* handle going up-right */
	if (arrow->steps[___][RGT] == '/') {
		arrow->row = row;
		pt->x -= XHSCALE(1);
		pt->y -= YHSCALE(1);
		follow_arrow_up_right(grid, arrow, pt);
		return;
	}

	/* handle going down-left */
	if (arrow->steps[DWN][___] == '/') {
		arrow->row = row;
		pt->x -= XHSCALE(1);
		pt->y -= YHSCALE(1);
		follow_arrow_down_left(grid, arrow, pt);
		return;
	}

	/* handle a skip */
	if (row < grid->nlines - 2 &&
	    col < grid->ncols - 2 &&
	    grid->lines[row + 1][col + 1] == '|' &&
	    grid->lines[row + 2][col + 2] == '\\'
	    ) {
		row += 2;
		col += 2;
		goto resume;
	}
}

/*****************************************************************************/
/*
 * follow an arrow junction or corner
 */
static void follow_arrow_junction(struct grid *grid, struct arrow *arrow, struct point *last)
{
	struct arrow *jarrow;
	int row = arrow->row, col = arrow->col;

	_debug("JUNCTION");

	map_arrow_steps(grid, arrow, row, col);
	_debug("- J1 step: %2d,%2d: '%c' [%s]",
	       row, col, arrow->steps[___][___], arrow->psteps);

	grid->markup[row][col] |= MARK_ARROW;

	/* main arrow follows the first line we find */
	if (arrow->steps[___][LFT] == '-')
		follow_arrow_left(grid, arrow, last);
	else if (arrow->steps[___][RGT] == '-')
		follow_arrow_right(grid, arrow, last);
	else if (arrow->steps[UP_][___] == '|')
		follow_arrow_up(grid, arrow, last);
	else if (arrow->steps[DWN][___] == '|')
		follow_arrow_down(grid, arrow, last);
	else
		return;

	/* may need secondary arrows */
	for (;;) {
		map_arrow_steps(grid, arrow, row, col);
		_debug("- J2 step: %2d,%2d: '%c' [%s]",
		       row, col, arrow->steps[___][___], arrow->psteps);

		if (arrow->steps[___][LFT] != '-' &&
		    arrow->steps[___][RGT] != '-' &&
		    arrow->steps[UP_][___] != '|' &&
		    arrow->steps[DWN][___] != '|')
			return;

		jarrow = xmalloc(sizeof(*jarrow));
		jarrow->row = row;
		jarrow->col = col;
		jarrow->end = A_OTHER;
		jarrow->shead = 0;
		jarrow->ehead = 0;
		jarrow->npoints = 1;
		jarrow->start.next = NULL;
		jarrow->start.x = last->x;
		jarrow->start.y = last->y;
		jarrow->link = grid->arrows;
		grid->arrows = jarrow;

		if (arrow->steps[___][LFT] == '-')
			follow_arrow_left(grid, jarrow, &jarrow->start);
		else if (arrow->steps[___][RGT] == '-')
			follow_arrow_right(grid, jarrow, &jarrow->start);
		else if (arrow->steps[UP_][___] == '|')
			follow_arrow_up(grid, jarrow, &jarrow->start);
		else if (arrow->steps[DWN][___] == '|')
			follow_arrow_down(grid, jarrow, &jarrow->start);
		else
			abort();
	}
}

/*****************************************************************************/
/*
 * shorten diagonals that join perpendicular horizontal/vertical lines by half
 * a notch at each end
 */
static void shorten_arrow_diagonals(struct grid *grid, struct arrow *arrow)
{
	struct point *pts[4], *first, *p;
	int loop;

	first = &arrow->start;

	/* we need to examine sets of three adjacent lines (four points) */
	p = first;
	for (loop = 0; loop < 4; loop++) {
		if (!p)
			return;
		pts[loop] = p;
		p = p->next;
	}

process:
	_debug("BEFORE (%d,%d) -> (%d,%d) -> (%d,%d) -> (%d,%d)",
	       pts[0]->x, pts[0]->y,
	       pts[1]->x, pts[1]->y,
	       pts[2]->x, pts[2]->y,
	       pts[3]->x, pts[3]->y);

	/* ditch the first line and shuffle down if not horiz/vert */
	if (pts[0]->x != pts[1]->x && pts[0]->y != pts[1]->y)
		goto move_to_next;

	/* or if the second line is not a diagonal */
	if (pts[1]->x == pts[2]->x || pts[1]->y == pts[2]->y)
		goto move_to_next;

	/* or if the third is not horiz/vert */
	if (pts[2]->x != pts[3]->x && pts[2]->y != pts[3]->y)
		goto move_to_next;

	/* first and third lines must be perpendicular */
	if ((pts[0]->x == pts[1]->x && pts[2]->x == pts[3]->x) ||
	    (pts[0]->y == pts[1]->y && pts[2]->y == pts[3]->y))
		goto move_to_next;

	/* extend the first and third lines into the diagonal */
	if (pts[1]->y == pts[0]->y && pts[1]->x < pts[0]->x) {
		/* horizontal left extension required */
		assert(pts[2]->x < pts[1]->x);
		assert(pts[3]->x == pts[2]->x);

		pts[1]->x -= XHSCALE(1);
		if (pts[3]->y < pts[2]->y)
			pts[2]->y += YHSCALE(1);
		else
			pts[2]->y -= YHSCALE(1);


	} else if (pts[1]->y == pts[0]->y && pts[1]->x > pts[0]->x) {
		/* horizontal right extension required */
		assert(pts[2]->x > pts[1]->x);
		assert(pts[3]->x == pts[2]->x);

		pts[1]->x += XHSCALE(1);
		if (pts[3]->y < pts[2]->y)
			pts[2]->y += YHSCALE(1);
		else
			pts[2]->y -= YHSCALE(1);

	} else if (pts[1]->x == pts[0]->x && pts[1]->y < pts[0]->y) {
		/* vertical up extension required */
		assert(pts[2]->y < pts[1]->y);
		assert(pts[3]->y == pts[2]->y);

		pts[1]->y -= YHSCALE(1);
		if (pts[3]->x < pts[2]->x)
			pts[2]->x += XHSCALE(1);
		else
			pts[2]->x -= XHSCALE(1);

	} else if (pts[1]->x == pts[0]->x && pts[1]->y > pts[0]->y) {
		/* vertical up extension required */
		assert(pts[2]->y > pts[1]->y);
		assert(pts[3]->y == pts[2]->y);

		pts[1]->y += YHSCALE(1);
		if (pts[3]->x < pts[2]->x)
			pts[2]->x += XHSCALE(1);
		else
			pts[2]->x -= XHSCALE(1);

	} else {
		abort();
	}

	_debug("AFTER  (%d,%d) -> (%d,%d) -> (%d,%d) -> (%d,%d)",
	       pts[0]->x, pts[0]->y,
	       pts[1]->x, pts[1]->y,
	       pts[2]->x, pts[2]->y,
	       pts[3]->x, pts[3]->y);

move_to_next:
	if (!p)
		return;
	pts[0] = pts[1];
	pts[1] = pts[2];
	pts[2] = pts[3];
	pts[3] = p;
	p = p->next;
	goto process;
}

/*****************************************************************************/
/*
 * free an arrow point
 */
static void free_arrow_point(struct point *point)
{
	if (point->next)
		free_arrow_point(point->next);
	if (point->branch)
		free_arrow(point->branch);
	free(point);
}

/*****************************************************************************/
/*
 * free an arrow
 */
static void free_arrow(struct arrow *arrow)
{
	if (arrow->start.next)
		free_arrow_point(arrow->start.next);
	free(arrow);
}

/*****************************************************************************/
/*
 * free up the arrows in a grid
 */
void free_arrows(struct grid *grid)
{
	struct arrow *arrow;

	while (grid->arrows) {
		arrow = grid->arrows;
		grid->arrows = arrow->link;

		free_arrow(arrow);
	}
}

/*****************************************************************************/
/*
 * render an arrow
 */
extern void render_arrow(const struct arrow *arrow, FILE *out)
{
	const struct point *pt;

	fprintf(out,
		"2"		/* object_code: polyline */
		" 1"		/* sub_type: polyline */
		" 0"		/* line_style: solid */
		" 1"		/* thickness */
		" 0"		/* pen_color: black */
		" 7"		/* fill_color: white */
		" 50"		/* depth */
		" -1"		/* pen_style - not used */
		" -1"		/* area_fill: no fill */
		" 0.000"	/* style_val (1/80 inch) */
		" 0"		/* join_style: mitre */
		" 0"		/* cap_style: butt */
		" -1"		/* radius */
		" %d"		/* forward_arrow */
		" %d"		/* backward_arrow */
		" %d"		/* npoints */
		"\n",
		arrow->ehead,
		arrow->shead,
		arrow->npoints);

	if (arrow->ehead)
		fprintf(out,
			"\t1"		/* arrow_type: closed triangle */
			" 1"		/* arrow_style: filled */
			" 0.00"		/* arrow_thickness */
			" 60.000"	/* arrow_width */
			" 120.000"	/* arrow_height */
			"\n");

	if (arrow->shead)
		fprintf(out,
			"\t1"		/* arrow_type: closed triangle */
			" 1"		/* arrow_style: filled */
			" 0.00"		/* arrow_thickness */
			" 60.000"	/* arrow_width */
			" 120.000"	/* arrow_height */
			"\n");

	fprintf(out, "\t");
	_debug("\nARROW");
	for (pt = &arrow->start; pt; pt = pt->next) {
		_debug("- to %d,%d", pt->x, pt->y);
		fprintf(out, " %d %d", pt->x,  pt->y);
	}
	fprintf(out, "\n");
}
