/* brace.c: ascii2xfig vertical brace parser and renderer
 *
 * Copyright (C) 2006 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include "defs.h"

static void extract_braces(struct grid *grid, int row, int col);

/*****************************************************************************/
/*
 * find all the vertical braces
 */
void find_braces(struct grid *grid)
{
	char **lines = grid->lines;
	int row, col;

	for (row = 0; row < grid->nlines; row++) {
		for (col = 0; col < grid->ncols; col++) {
			if (grid->markup[row][col])
				continue;
			if (lines[row][col] == '}')
				extract_braces(grid, row, col);
		}
	}
}

/*****************************************************************************/
/*
 * extract a vertical brace
 * - a vertical brace is a vertical line of closing braces with nothing inbetween
 */
static void extract_braces(struct grid *grid, int row, int col)
{
	struct vline *brace;

	brace = xmalloc(sizeof(*brace));
	brace->top = row;
	brace->bottom = row;
	brace->col = col;
	brace->link = grid->vbraces;
	grid->vbraces = brace;

	for (row++; grid->lines[row][col]; row++) {
		if (grid->lines[row][col] != '}')
			break;

		brace->bottom = row;
	}

	/* prevent the braces from being seen again */
	for (row = brace->top; row <= brace->bottom; row++)
		grid->markup[row][col] |= MARK_TEXT;
}

/*****************************************************************************/
/*
 * free up the vertical braces in a grid
 */
void free_braces(struct grid *grid)
{
	struct vline *brace;

	while (grid->vbraces) {
		brace = grid->vbraces;
		grid->vbraces = brace->link;
		free(brace);
	}
}

/*****************************************************************************/
/*
 * render a brace as a pair of vertical lines
 */
void render_brace(const struct vline *brace, FILE *out)
{
	int x, xc, xq = XSCALE(1) / 4;
	int y0, y1, ym, yat, yab, yq = YSCALE(1) / 4;

	kdebug("BRACES: (%d,%d) -> (%d,%d)",
	       brace->top, brace->col, brace->bottom, brace->col);

#if 1
	x = XSCALE(brace->col);

	/* top and bottom of brace */
	y0 = YSCALE(brace->top) - yq;
	y1 = YSCALE(brace->bottom + 1) + yq;

	/* arrow in middle of brace */
	ym = (y0 + y1) / 2;
	yat = ym - YHSCALE(1);
	yab = ym + YHSCALE(1);

	/* curve of brace */
	xc = x + xq * 2;

	/* draw top curve */
	fprintf(out,
		"3"		/* object_code: spline */
		" 2"		/* sub_type: open interpolated spline */
		" 0"		/* line_style: solid */
		" 2"		/* thickness */
		" 0"		/* pen_color: black */
		" 7"		/* fill_color: white */
		" 52"		/* depth */
		" -1"		/* pen_style - not used */
		" -1"		/* area_fill: no fill */
		" 0.000"	/* style_val (1/80 inch) */
		" 0"		/* cap_style: butt */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 3"		/* npoints */
		"\n");

	fprintf(out, " %d %d", x, y0);
	fprintf(out, " %d %d", xc, (y0 + yat) / 2);
	fprintf(out, " %d %d", x, yat);
	fprintf(out, "\n");

	fprintf(out, "\t 0.000 -1.000 0.000\n");

	/* draw arrow */
	fprintf(out,
		"2"		/* object_code: polyline */
		" 1"		/* sub_type: polyline */
		" 0"		/* line_style: solid */
		" 2"		/* thickness */
		" 0"		/* pen_color: black */
		" 7"		/* fill_color: white */
		" 52"		/* depth */
		" -1"		/* pen_style - not used */
		" -1"		/* area_fill: no fill */
		" 0.000"	/* style_val (1/80 inch) */
		" 0"		/* join_style: mitre */
		" 0"		/* cap_style: butt */
		" -1"		/* radius */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 3"		/* npoints */
		"\n");

	fprintf(out, "\t");
	fprintf(out, " %d %d", x, yat);
	fprintf(out, " %d %d", x + XHSCALE(2), ym);
	fprintf(out, " %d %d", x, yab);
	fprintf(out, "\n");

	/* draw bottom curve */
	fprintf(out,
		"3"		/* object_code: spline */
		" 2"		/* sub_type: open interpolated spline */
		" 0"		/* line_style: solid */
		" 2"		/* thickness */
		" 0"		/* pen_color: black */
		" 7"		/* fill_color: white */
		" 52"		/* depth */
		" -1"		/* pen_style - not used */
		" -1"		/* area_fill: no fill */
		" 0.000"	/* style_val (1/80 inch) */
		" 0"		/* cap_style: butt */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 3"		/* npoints */
		"\n");

	fprintf(out, " %d %d", x, yab);
	fprintf(out, " %d %d", xc, (yab + y1) / 2);
	fprintf(out, " %d %d", x, y1);
	fprintf(out, "\n");

	fprintf(out, "\t 0.000 -1.000 0.000\n");

#else
	fprintf(out,
		"2"		/* object_code: polyline */
		" 1"		/* sub_type: polyline */
		" 0"		/* line_style: solid */
		" 1"		/* thickness */
		" 0"		/* pen_color: black */
		" 7"		/* fill_color: white */
		" 52"		/* depth */
		" -1"		/* pen_style - not used */
		" -1"		/* area_fill: no fill */
		" 0.000"	/* style_val (1/80 inch) */
		" 0"		/* join_style: mitre */
		" 0"		/* cap_style: butt */
		" -1"		/* radius */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 4"		/* npoints */
		"\n");

	fprintf(out, "\t");
	fprintf(out, " %d %d", XSCALEH(brace->col - 1), YSCALE(brace->top) - yq);
	fprintf(out, " %d %d", XSCALEH(brace->col)    , YSCALE(brace->top) - yq);
	fprintf(out, " %d %d", XSCALEH(brace->col)    , YSCALE(brace->bottom + 1) + yq);
	fprintf(out, " %d %d", XSCALEH(brace->col - 1), YSCALE(brace->bottom + 1) + yq);
	fprintf(out, "\n");

	fprintf(out,
		"2"		/* object_code: polyline */
		" 1"		/* sub_type: polyline */
		" 0"		/* line_style: solid */
		" 2"		/* thickness */
		" 0"		/* pen_color: black */
		" 7"		/* fill_color: white */
		" 52"		/* depth */
		" -1"		/* pen_style - not used */
		" -1"		/* area_fill: no fill */
		" 0.000"	/* style_val (1/80 inch) */
		" 0"		/* join_style: mitre */
		" 0"		/* cap_style: butt */
		" 60"		/* radius */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 2"		/* npoints */
		"\n");

	fprintf(out, "\t");
	fprintf(out, " %d %d", XSCALEH(brace->col) - xq, YSCALE(brace->top) - yq);
	fprintf(out, " %d %d", XSCALEH(brace->col) - xq, YSCALE(brace->bottom + 1) + yq);
	fprintf(out, "\n");
#endif
}
