/* defs.h: ascii2xfig definitions
 *
 * Copyright (C) 2006 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <limits.h>
#include <assert.h>

#define SYSERROR(X, Y) do { if (X) { perror(Y); exit(1); } } while(0)

struct arrow;
struct box;
struct sequence;
struct text;
struct textbox;
struct vlines;

struct grid {
	char		**lines;	/* lines in grid */
	unsigned char	**markup;	/* lines in markup grid */
	int		ncols;		/* length of a line */
	int		nlines;		/* number of lines */
	struct box	*boxes;		/* boxes found in this grid */
	struct sequence	*sequences;	/* sequences found in this grid */
	struct arrow	*arrows;	/* arrows */
	struct textbox	*texts;		/* blocks of text */
	struct vline	*vdots;		/* vertical dotted lines */
	struct vline	*vbraces;	/* vertical braces */
	char		*_lines[0];
};

#define MARK_BOX	0x01U		/* flag to indicate box chars */
#define MARK_SEQ	0x02U		/* flag to indicate sequence chars */
#define MARK_TEXT	0x04U		/* flag to indicate text chars */
#define MARK_ARROW	0x08U		/* flag to indicate arrow chars */
#define MARK_SEEN	0x80U		/* flag to prevent recursion */

struct box {
	int		top;		/* start row */
	int		left;		/* start column */
	int		bottom;		/* end row */
	int		right;		/* end column */

	struct box	*link;		/* next box in grid list */
	struct text	*text;		/* bits of text inside the box */
};

struct text {
	int		len;
	int		row;
	struct text	*link;		/* link in list from parent */
	char		text[0];
};

struct textbox {
	int		col;		/* column in which text starts */
	struct text	*rows;
	struct textbox	*link;		/* link in grid list */
};

struct element {
	struct element	*link;		/* link in sequence list */
	union {
		struct box	*box;	/* box inside the sequence */
		struct text	*text;	/* text inside the sequence */
	};
	enum {
		ELEM_BOX,	/* box */
		ELEM_TEXT,	/* text insert */
	} type;
	int		top;		/* top of element */
	int		left;		/* left hand end of element */
};

struct sequence {
	int		top;		/* start row */
	int		left;		/* start column */
	int		bottom;		/* end row */
	int		right;		/* end column */

	struct sequence	*link;		/* next box in grid list */
	struct element	*elements;	/* elements in a sequence */
};

struct point {
	struct point	*next;		/* next point */
	struct arrow	*branch;	/* junction branch */
	int		skip;
	int		x;
	int		y;
};

typedef enum {
	A_LEFT,
	A_RIGHT,
	A_UP,
	A_DOWN,
	A_UP_LEFT,
	A_UP_RIGHT,
	A_DOWN_LEFT,
	A_DOWN_RIGHT,
	A_OTHER,
} arrow_t;

struct arrow {
	int		row;		/* current row */
	int		col;		/* current col */
	int		npoints;	/* number of segments */
	arrow_t		end;		/* direction to follow arrow in */
	char		steps[3][3];	/* following grid */
	char		psteps[12];	/* text representation of arrow surrounds */
	char		shead;		/* T if arrowhead at start */
	char		ehead;		/* T if arrowhead at end */
	struct arrow	*link;		/* next arrow in grid list */
	struct point	start;		/* first point of this arrow */
};

struct vline {
	int		top;		/* top row of line */
	int		bottom;		/* bottom row of line */
	int		col;		/* column in which line resides */
	struct vline	*link;		/* next vertical dotted line in grid */
};

/* arrow.c */
extern void find_arrows(struct grid *grid);
extern void free_arrows(struct grid *grid);
extern void render_arrow(const struct arrow *arrow, FILE *out);

/* box.c */
extern void find_boxes(struct grid *grid);
extern void free_boxes(struct grid *grid);
extern void free_box(struct box *box);
extern void render_box(const struct box *box, FILE *out);

/* brace.c */
extern void find_braces(struct grid *grid);
extern void free_braces(struct grid *grid);
extern void render_brace(const struct vline *brace, FILE *out);

/* sequence.c */
extern void find_sequences(struct grid *grid);
extern void free_sequences(struct grid *grid);
extern void render_sequence(const struct sequence *seq, FILE *out);

/* text.c */
extern void find_texts(struct grid *grid);
extern void free_texts(struct grid *grid);
extern void render_text(const struct textbox *tb, FILE *out);

/* vdots.c */
extern void find_vdots(struct grid *grid);
extern void free_vdots(struct grid *grid);
extern void render_vdot(const struct vline *vdot, FILE *out);

/* miscellany */
static inline void *xmalloc(size_t n)
{
	void *p = malloc(n);
	SYSERROR(!p, "malloc");
	return p;
}

static inline void free_text(struct text *text)
{
	free(text);
}


#define kdebug(FMT, ...) fprintf(stderr, FMT"\n" ,##__VA_ARGS__ )

#ifdef KDEBUG
#define _debug kdebug
#else
static inline __attribute__((format(printf,1,2)))
void  _debug(const char *fmt, ...) {}
#endif

/* scaling */
#define SBASE	120		/* main scaling factor (applies to X and Y both) */
#define XSBASE	(SBASE * 3 / 4)	/* ascii-char box X-size scaling factor */
#define YSBASE	SBASE		/* ascii-char box Y-size scaling factor */

#define XSCALE(X)	((int) ((X) * XSBASE))
#define XHSCALE(X)	((int) ((X) * (XSBASE / 2)))
#define XSCALEH(X)	((int) ((X) * XSBASE + XSBASE / 2))
#define XSCALEHF(X,N,D)	((int) ((((X) * (D) + (N)) * XSBASE) / (D) + XSBASE / 2))

#define YSCALE(X)	((int) ((X) * YSBASE))
#define YHSCALE(X)	((int) ((X) * (YSBASE / 2)))
#define YSCALEH(X)	((int) ((X) * YSBASE + YSBASE / 2))
#define YSCALEHF(X,N,D)	((int) ((((X) * (D) + (N)) * YSBASE) / (D) + YSBASE / 2))
