/* ascii2xfig.c: Convert an ASCII diagram into xfig
 *
 * Copyright (C) 2006 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include "defs.h"

static struct grid *read_ascii_file(const char *filename);
static void dump_grid(struct grid *grid) __attribute__((used));
static void render(const struct grid *grid, FILE *out);

/*****************************************************************************/
/*
 *
 */
int main(int argc, char *argv[])
{
	struct grid *grid;

	for (argv++; *argv; argv++) {
		grid = read_ascii_file(*argv);

		find_boxes(grid);
		find_sequences(grid);
		find_texts(grid);
		find_arrows(grid);
		find_vdots(grid);
		find_braces(grid);

		//dump_grid(grid);
		render(grid, stdout);

		/* free up the memory */
		free_braces(grid);
		free_vdots(grid);
		free_arrows(grid);
		free_texts(grid);
		free_sequences(grid);
		free_boxes(grid);
		free(grid);
	}

	return 0;
}

/*****************************************************************************/
/*
 * read the entire ascii file in
 * - need to pad out TAB chars
 */
static struct grid *read_ascii_file(const char *filename)
{
	struct grid *grid;
	struct stat st;
	size_t size;
	char *_tbuf, *tbuf, *buffer, *p, *q, **lines, **pp;
	int fd, col, line, indent, ncols, nlines;

	/* read the whole file */
	fd = open(filename, O_RDONLY);
	SYSERROR(fd < 0, filename);

	SYSERROR(fstat(fd, &st) < 0, filename);

	if (st.st_size == 0) {
		fprintf(stderr, "%s: Input file is empty\n", filename);
		exit(1);
	}

	tbuf = _tbuf = xmalloc(st.st_size + 2);

	size = st.st_size;
	size = read(fd, tbuf, size);
	SYSERROR(size < 0, filename);
	SYSERROR(close(fd) < 0, filename);

	if (tbuf[size - 1] != '\n')
		tbuf[size++] = '\n';

	tbuf[size] = 0;

	/* delete leading and trailing blank lines */
	while (*tbuf == '\n') {
		tbuf++;
		size--;
	}

	if (size < 0)
		goto blank_file;

	while (tbuf[size - 1] == '\n' &&
	       tbuf[size - 2] == '\n')
		size--;

	tbuf[size] = 0;

	/* count the number of lines and work out max line length */
	nlines = 0;
	ncols = 0;
	p = tbuf;

	do {
		nlines++;
		col = 0;

		for (; *p != '\n'; p++) {
			col++;
			/* allow for tab expansion */
			if (*p == '\t')
				col = (col + 7) & ~7;
		}

		if (col > ncols)
			ncols = col;
		p++;
	} while (*p);

	/* allocate a square grid in which to place the data and a square grid
	 * to hold markups
	 * - allocate a zeroised row above and below
	 * - allocate a zeroised column to either side
	 */
	grid = xmalloc(sizeof(*grid) +
		       sizeof(char *) * (nlines + 3) * 2 +
		       (nlines + 2) * (ncols + 2) * 2);

	grid->nlines = nlines;
	grid->ncols = ncols;
	grid->lines = grid->_lines;
	grid->markup = (unsigned char **) grid->lines + (nlines + 3);
	grid->boxes = NULL;
	grid->sequences = NULL;
	grid->arrows = NULL;
	grid->texts = NULL;
	grid->vdots = NULL;
	grid->vbraces = NULL;

	buffer = (char *) grid;
	buffer += sizeof(*grid) + sizeof(char *) * (nlines + 3) * 2;

	/* fill in the row pointers */
	q = buffer;
	pp = grid->lines;

	for (line = nlines + 2; line > 0; line--) {
		*q++ = '\0';
		*pp++ = q;
		q += ncols + 1;
	}
	*pp++ = NULL;
	grid->lines++;

	for (line = nlines + 2; line > 0; line--) {
		*q++ = '\0';
		*pp++ = q;
		q += ncols + 1;
	}
	*pp++ = NULL;
	grid->markup++;

	assert(grid->lines[grid->nlines + 1] == NULL);
	assert(grid->markup[grid->nlines + 1] == NULL);

	/* transfer the data to the grid */
	p = tbuf;

	lines = grid->lines;
	line = 0;

	do {
		q = lines[line];
		col = 0;

		while (*p != '\n') {
			/* expand tabs */
			if (*p == '\t') {
				p++;
				do {
					*q++ = ' ';
					col++;
				} while (col & 7);
			} else {
				*q++ = *p++;
				col++;
			}
		}

		p++;
		assert(q - lines[line] <= ncols);

		while (q - lines[line] < ncols)
			*q++ = ' ';

		*q = '\0';
		line++;
	} while (*p);

	assert(lines[grid->nlines + 1] == NULL);

	/* remove leading and trailing blank lines and common indentation */
	indent = INT_MAX;

	for (line = 0; line < grid->nlines; line++) {
		col = strspn(lines[line], " ");
		if (col == ncols) {
			if (line == 0) {
				lines++;
				grid->lines++;
				grid->nlines--;
				line--;
				continue;
			}
		}

		if (col < indent)
			indent = col;
	}

	assert(lines[grid->nlines + 1] == NULL);

	for (line = 0; line < grid->nlines; line++) {
		lines[line] += indent;
		lines[line][-1] = '\0';
	}
	grid->ncols -= indent;

	if (grid->nlines == 0)
		goto blank_file;

	while (strspn(lines[grid->nlines - 1], " ") == grid->ncols)
		grid->nlines--;

	lines[grid->nlines + 1] = NULL;

	/* initialise the data grid buffer lines */
	memset(lines[-1] - 1, 0, grid->ncols + 1);
	memset(lines[grid->nlines] - 1, 0, grid->ncols + 1);

	/* initialise the markup grid to match the data grid */
	lines = (char **) grid->markup;
	p = lines[0];
	for (line = 0; line < grid->nlines; line++) {
		q = lines[line] = p + line * (grid->ncols + 1);
		memset(q, 0, grid->ncols + 1);
	}
	lines[grid->nlines + 1] = NULL;

	/* initialise the markup grid buffer lines */
	memset(lines[-1] - 1, 0, grid->ncols + 1);
	memset(lines[grid->nlines] - 1, 0, grid->ncols + 1);

	*q = '\0';
	free(_tbuf);
	return grid;

blank_file:
	fprintf(stderr, "%s: Input file is blank\n", filename);
	exit(1);
}

/*****************************************************************************/
/*
 *
 */
static void dump_grid(struct grid *grid)
{
	FILE *dout = stderr;
	char **lines;
	int row, col;

	lines = grid->lines;

	for (row = 0; row < grid->nlines; row++) {
		for (col = 0; col < grid->ncols; col++) {
			if (grid->markup[row][col])
				fprintf(dout, "\e[4%cm%c\e[m",
					'0' + (grid->markup[row][col] & 7),
					lines[row][col]);
			else
				fprintf(dout, "%c", lines[row][col]);
		}
		fputc('\n', dout);
	}
}

/*****************************************************************************/
/*
 * render a grid for debugging purposes
 */
static __attribute__((used))
void render_grid(const struct grid *grid, FILE *out)
{
	int row, col;

	for (row = 0; row <= grid->nlines; row++) {
		fprintf(out,
			"2"		/* object_code: polyline */
			" 1"		/* sub_type: polyline */
			" 2"		/* line_style: dotted */
			" 1"		/* thickness */
			" 6"		/* pen_color: yellow */
			" 7"		/* fill_color: white */
			" 100"		/* depth */
			" -1"		/* pen_style - not used */
			" -1"		/* area_fill: no fill */
			" 1.000"	/* style_val (1/80 inch) */
			" 0"		/* join_style: mitre */
			" 0"		/* cap_style: butt */
			" -1"		/* radius */
			" 0"		/* forward_arrow */
			" 0"		/* backward_arrow */
			" 2"		/* npoints */
			"\n");

		fprintf(out, "\t");
		fprintf(out, " %d %d", XSCALE(0),	    YSCALE(row));
		fprintf(out, " %d %d", XSCALE(grid->ncols), YSCALE(row));
		fprintf(out, "\n");
	}

	for (col = 0; col <= grid->ncols; col++) {
		fprintf(out,
			"2"		/* object_code: polyline */
			" 1"		/* sub_type: polyline */
			" 2"		/* line_style: dotted */
			" 1"		/* thickness */
			" 6"		/* pen_color: yellow */
			" 7"		/* fill_color: white */
			" 100"		/* depth */
			" -1"		/* pen_style - not used */
			" -1"		/* area_fill: no fill */
			" 1.000"	/* style_val (1/80 inch) */
			" 0"		/* join_style: mitre */
			" 0"		/* cap_style: butt */
			" -1"		/* radius */
			" 0"		/* forward_arrow */
			" 0"		/* backward_arrow */
			" 2"		/* npoints */
			"\n");

		fprintf(out, "\t");
		fprintf(out, " %d %d", XSCALE(col), YSCALE(0));
		fprintf(out, " %d %d", XSCALE(col), YSCALE(grid->nlines));
		fprintf(out, "\n");
	}
}

/*****************************************************************************/
/*
 * attempt to render to an XFIG file
 */
static void render(const struct grid *grid, FILE *out)
{
	const struct sequence *seq;
	const struct textbox *tb;
	const struct arrow *arrow;
	const struct vline *vl;
	const struct box *box;

	fprintf(out, "#FIG 3.2\n");
	fprintf(out, "Landscape\n");	/* orientation */
	fprintf(out, "Center\n");	/* justification */
	fprintf(out, "Inches\n");	/* units */
	fprintf(out, "A4\n");		/* paper size */
	fprintf(out, "100.00\n");	/* magnification */
	fprintf(out, "Single\n");	/* multiple page */
	fprintf(out, "-2\n");		/* transparent colour */
	fprintf(out, "%d 2\n", SBASE * 10); /* resolution & coordinate system: upper left*/

	/* colour definitions */
	fprintf(out,	/* 32: light grey */
		"0"			/* object_code: Color Pseudo-objects */
		" 32"			/* color_number */
		" #efefef"		/* hex RGB colour */
		"\n");

	fprintf(out,	/* 33: dark grey */
		"0"			/* object_code: Color Pseudo-objects */
		" 33"			/* color_number */
		" #7f7f7f"		/* hex RGB colour */
		"\n");

	//render_grid(grid, out);

	for (vl = grid->vdots; vl; vl = vl->link)
		render_vdot(vl, out);

	for (box = grid->boxes; box; box = box->link)
		render_box(box, out);

	for (seq = grid->sequences; seq; seq = seq->link)
		render_sequence(seq, out);

	for (vl = grid->vbraces; vl; vl = vl->link)
		render_brace(vl, out);

	for (arrow = grid->arrows; arrow; arrow = arrow->link)
		render_arrow(arrow, out);

	for (tb = grid->texts; tb; tb = tb->link)
		render_text(tb, out);
}
