/* sequence.c: ascii2xfig vertical sequence parser and renderer
 *
 * Copyright (C) 2006 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include "defs.h"

static void extract_sequence(struct grid *grid, int row, int col);
static void extract_sequence_box(struct grid *grid, struct sequence *seq,
				 struct element ***_elem, int *_row);
static void extract_sequence_text(struct grid *grid, struct sequence *seq,
				  struct element ***_elem, int *_row);

/*****************************************************************************/
/*
 * find all the sequences in the diagram
 */
void find_sequences(struct grid *grid)
{
	char **lines = grid->lines;
	int row, col;

	/* search on all but the bottom row and RHS column */
	for (row = 0; row < grid->nlines - 2; row++) {
		for (col = 0; col < grid->ncols - 2; col++) {
			if (lines[row][col] == ':')
				extract_sequence(grid, row, col);
		}
	}
}

/*****************************************************************************/
/*
 * attempt to extract a vertical sequence from the diagram at a given row and
 * column
 * - a vertical sequence is an open rectangle with a ':' at each corner
 * - there must be a space (or the top of diagram) above each top colon
 * - there must be a space (or the bottom of diagram) below each top colon
 * - a vseq is a pair of vertical lines of colons with boxes and text overlaid
 * - the sequence goes left and down from the current position
 */
static void extract_sequence(struct grid *grid, int row, int col)
{
	struct sequence seq, *pseq;
	struct element **ppelem;
	char **lines = grid->lines;
	int n;

	seq.top = row;
	seq.left = col;

	/* if there's a non-space char above then it's not a sequence */
	if (row > 0 && lines[row - 1][col - 1] != ' ')
		return;

	/* look for the right-top colon */
	for (n = col + 1; n < grid->ncols; n++) {
		if (grid->markup[seq.top][n])
			return;

		switch (lines[seq.top][n]) {
		case ' ':
			continue;
		case ':':
			goto found_rtop;
		case '^':
		case '|':
		case 'V':
			continue; /* vertical arrow */
		default:
			return;
		}
	}

	return;

found_rtop:
	seq.right = n;
	_debug("maybe SEQ at (%d,%d) to (?,%d)",
	       seq.top, seq.left, seq.right);

	/* if there's a non-space char above then it's not a sequence */
	if (row > 0 && lines[row - 1][n] != ' ')
		return;

	/* trace down both sides of the sequence */
	for (n = seq.top + 1; n < grid->nlines; n++) {
		char l, r;

		l = lines[n][seq.left];
		r = lines[n][seq.right];

		if (l == ' ' && r == ' ')
			break;

		if (l == ':' && r == ':') {
			seq.bottom = n;
			continue;
		}

		if ((l == '+' && r == '+') ||
		    (l == '|' && r == '|')
		    ) {
			if (grid->markup[n][seq.left] & MARK_BOX)
				continue;
		}

		if (isalnum(l) && isalnum(r))
			continue;

		_debug("Reject: '%c' & '%c'", l, r);
		return;
	}

	if (seq.bottom == 0)
		return;

	_debug("SEQ at (%d,%d) to (%d,%d)",
	       seq.top, seq.left, seq.bottom, seq.right);

	pseq = xmalloc(sizeof(*pseq));
	pseq->top = seq.top;
	pseq->left = seq.left;
	pseq->bottom = seq.bottom;
	pseq->right = seq.right;
	pseq->link = grid->sequences;
	grid->sequences = pseq;

	/* extract the elements of a sequence */
	ppelem = &pseq->elements;
	for (n = seq.top; n <= seq.bottom;) {
		grid->markup[n][seq.left] |= MARK_SEQ;
		grid->markup[n][seq.right] |= MARK_SEQ;

		switch (lines[n][seq.left]) {
		case ':':
			n++;
			continue;
		case '+':
			extract_sequence_box(grid, pseq, &ppelem, &n);
			break;
		case '0' ... '9':
		case 'a' ... 'z':
		case 'A' ... 'Z':
		case '_':
			extract_sequence_text(grid, pseq, &ppelem, &n);
			break;
		default:
			fprintf(stderr, "Unexpected char '%c'\n", lines[n][seq.left]);
			abort();
		}
	}
	*ppelem = NULL;
}

/*****************************************************************************/
/*
 * add a box sequence element
 */
static void extract_sequence_box(struct grid *grid, struct sequence *seq,
				 struct element ***_elem, int *_row)
{
	struct element *elem;
	struct box *box, **ppbox;
	int row = *_row;

	/* find the box */
	for (ppbox = &grid->boxes; *ppbox; ppbox = &(*ppbox)->link) {
		box = *ppbox;
		if (box->left == seq->left && box->top == row)
			goto found_box;
	}

	/* must've been the bottom-left corner of a box */
	*_row += 1;
	return;

found_box:
	elem = xmalloc(sizeof(*elem));
	elem->type = ELEM_BOX;
	elem->top = row;
	elem->box = box;
	elem->left = box->left;
	**_elem = elem;
	*_elem = &elem->link;

	*_row = box->bottom;

	/* remove the box from the master list */
	*ppbox = box->link;
	box->link = NULL;
}

/*****************************************************************************/
/*
 * add a text insert sequence element
 */
static void extract_sequence_text(struct grid *grid, struct sequence *seq,
				  struct element ***_elem, int *_row)
{
	struct element *elem;
	struct text *text;
	char **lines = grid->lines;
	int row, from, to, len;

	row = *_row;
	*_row = row + 1;

	elem = xmalloc(sizeof(*elem));
	elem->type = ELEM_TEXT;
	elem->top = row;
	**_elem = elem;
	*_elem = &elem->link;

	/* find the left end */
	for (from = seq->left; from >= 0; from--)
		if (!isalnum(lines[row][from]))
			break;
	from++;
	elem->left = from;

	/* find the right end */
	for (to = seq->right; to < grid->ncols; to++)
		if (!isalnum(lines[row][to]))
			break;

	len = to - from;
	_debug("TEXT: '%*.*s'", len, len, lines[row] + from);

	text = xmalloc(sizeof(*text) + len + 1);
	text->len = len;
	text->row = row;
	memcpy(text->text, lines[row] + from, len);
	text->text[len] = '\0';
	text->link = NULL;
	elem->text = text;

	/* mark the text */
	for (; from < to; from++)
		grid->markup[row][from] |= MARK_SEQ;
}

/*****************************************************************************/
/*
 * free up the sequences in a grid
 */
void free_sequences(struct grid *grid)
{
	struct sequence *seq;
	struct element *elem;

	while (grid->sequences) {
		seq = grid->sequences;
		grid->sequences = seq->link;

		while (seq->elements) {
			elem = seq->elements;
			seq->elements = elem->link;

			switch (elem->type) {
			case ELEM_BOX:
				free_box(elem->box);
				break;
			case ELEM_TEXT:
				free_text(elem->text);
				break;
			}

			free(elem);
		}

		free(seq);
	}
}

/*****************************************************************************/
/*
 * render a gap in a sequence as a pair of vertical grey lines, each with a
 * zig-zag "skip" symbol embedded in it
 */
static void render_sequence_gap(const struct sequence *seq, int from, int to, int col, FILE *out)
{
	fprintf(out,
		"2"		/* object_code: polyline */
		" 1"		/* sub_type: polyline */
		" 0"		/* line_style: solid */
		" 1"		/* thickness */
		" 33"		/* pen_color: light grey (user defined) */
		" 7"		/* fill_color: white */
		" 50"		/* depth */
		" -1"		/* pen_style - not used */
		" -1"		/* area_fill: no fill */
		" 1.000"	/* style_val (1/80 inch) */
		" 0"		/* join_style: mitre */
		" 0"		/* cap_style: butt */
		" -1"		/* radius */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 6"		/* npoints */
		"\n");

	fprintf(out, "\t");
	fprintf(out, " %d %d", XSCALEH(col),  YSCALEH(from));
	fprintf(out, " %d %d", XSCALEH(col),  YSCALEHF(from, 6, 20));
	fprintf(out, " %d %d", XSCALEHF(col, -5, 20),  YSCALEHF(from, 8, 20));
	fprintf(out, " %d %d", XSCALEHF(col, +5, 20),  YSCALEHF(from, 11, 20));
	fprintf(out, " %d %d", XSCALEH(col),  YSCALEHF(from, 14, 20));
	fprintf(out, " %d %d", XSCALEH(col),  YSCALEH(to));
	fprintf(out, "\n");


#if 0
	fprintf(out,
		"2"		/* object_code: polyline */
		" 1"		/* sub_type: polyline */
		" 2"		/* line_style: dotted */
		" 1"		/* thickness */
		" 0"		/* pen_color: black */
		" 7"		/* fill_color: white */
		" 50"		/* depth */
		" -1"		/* pen_style - not used */
		" -1"		/* area_fill: no fill */
		" 1.000"	/* style_val (1/80 inch) */
		" 0"		/* join_style: mitre */
		" 0"		/* cap_style: butt */
		" -1"		/* radius */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 2"		/* npoints */
		"\n");

	fprintf(out, "\t");
	fprintf(out, " %d %d", XSCALEH(col),  YSCALEH(from));
	fprintf(out, " %d %d", XSCALEH(col),  YSCALEH(to));
	fprintf(out, "\n");
#endif
}

/*****************************************************************************/
/*
 * render a text insert in a sequence
 */
void render_sequence_text(const struct element *elem, int centre, FILE *out)
{
	fprintf(out,
		"4"		/* object: text */
		" 1"		/* sub_type: centre justified */
		" 0"		/* color: black */
		" 50"		/* depth */
		" -1"		/* pen_style - not used */
		" 12"		/* font: courier */
		" 10"		/* font_size (pts) */
		" 0.000"	/* angle (radians) */
		" 4"		/* font flags: PS font */
		" %d"		/* height */
		" %d"		/* length */
		" %d"		/* x */
		" %d"		/* y */
		" %s\\001"	/* string */
		"\n",
		2,
		XSCALE(elem->text->len),
		centre,
		YSCALEH(elem->text->row) + (int) YHSCALE(1.5),
		elem->text->text
		);
}

/*****************************************************************************/
/*
 * render a sequence as XFIG
 */
void render_sequence(const struct sequence *seq, FILE *out)
{
	const struct element *elem;
	int row, centre;

	row = seq->top;

	centre = XSCALEH(seq->left) + XHSCALE(seq->right - seq->left);

	/* hatched backdrop to sequence */
	fprintf(out,
		"2"		/* object_code: polyline */
		" 2"		/* sub_type: box */
		" 0"		/* line_style: solid */
		" 0"		/* thickness */
		" 32"		/* pen_color: light grey (user defined) */
		" 7"		/* fill_color: white */
		" 51"		/* depth */
		" -1"		/* pen_style - not used */
		" 45"		/* area_fill: light grey fill */
		" 0.000"	/* style_val (1/80 inch) */
		" 0"		/* join_style: mitre */
		" 0"		/* cap_style: butt */
		" -1"		/* radius */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 5"		/* npoints */
		"\n");

	fprintf(out, "\t");
	fprintf(out, " %d %d", XSCALEH(seq->left),  YSCALEH(seq->top));
	fprintf(out, " %d %d", XSCALEH(seq->left),  YSCALEH(seq->bottom));
	fprintf(out, " %d %d", XSCALEH(seq->right), YSCALEH(seq->bottom));
	fprintf(out, " %d %d", XSCALEH(seq->right), YSCALEH(seq->top));
	fprintf(out, " %d %d", XSCALEH(seq->left),  YSCALEH(seq->top));
	fprintf(out, "\n");

	/* intersperse elements with dotted lines */
	for (elem = seq->elements; elem; elem = elem->link) {
		if (row < elem->top) {
			render_sequence_gap(seq, row, elem->top, seq->left, out);
			render_sequence_gap(seq, row, elem->top, seq->right, out);
			row = elem->top;
		}

		switch (elem->type) {
		case ELEM_BOX:
			render_box(elem->box, out);
			row = elem->box->bottom;
			break;
		case ELEM_TEXT:
			render_sequence_text(elem, centre, out);
			row++;
			break;
		}
	}

	/* render the bottom bit */
	render_sequence_gap(seq, row, seq->bottom, seq->left, out);
	render_sequence_gap(seq, row, seq->bottom, seq->right, out);
}
