/* vdots.c: ascii2xfig vertical dotted line parser and renderer
 *
 * Copyright (C) 2006 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include "defs.h"

static void extract_vdots(struct grid *grid, int row, int col);

/*****************************************************************************/
/*
 * find all the vertical dotted lines
 */
void find_vdots(struct grid *grid)
{
	char **lines = grid->lines;
	int row, col;

	for (row = 0; row < grid->nlines; row++) {
		for (col = 0; col < grid->ncols; col++) {
			if (grid->markup[row][col])
				continue;
			if (lines[row][col] == ':')
				extract_vdots(grid, row, col);
		}
	}
}

/*****************************************************************************/
/*
 * extract a vertical dotted line
 * - a vertical dotted line is a vertical line of colons with no spaces inbetween
 */
static void extract_vdots(struct grid *grid, int row, int col)
{
	struct vline *vdot;

	vdot = xmalloc(sizeof(*vdot));
	vdot->top = row;
	vdot->bottom = row;
	vdot->col = col;
	vdot->link = grid->vdots;
	grid->vdots = vdot;

	for (row++; grid->lines[row][col]; row++) {
		switch (grid->lines[row][col]) {
		case ' ':
			goto found_bottom;
		case ':':
			vdot->bottom = row;
		default:
			continue;
		}
	}
found_bottom:

	/* prevent the colons from being seen again */
	for (row = vdot->top; row <= vdot->bottom; row++)
		if (grid->lines[row][col] == ':')
			grid->markup[row][col] |= MARK_TEXT;
}

/*****************************************************************************/
/*
 * free up the vertical dotted lines in a grid
 */
void free_vdots(struct grid *grid)
{
	struct vline *vdot;

	while (grid->vdots) {
		vdot = grid->vdots;
		grid->vdots = vdot->link;
		free(vdot);
	}
}

/*****************************************************************************/
/*
 * render a line of vertical dots
 */
void render_vdot(const struct vline *vdot, FILE *out)
{
	fprintf(out,
		"2"		/* object_code: polyline */
		" 1"		/* sub_type: polyline */
		" 2"		/* line_style: dotted */
		" 1"		/* thickness */
		" 33"		/* pen_color: dark grey */
		" 7"		/* fill_color: white */
		" 52"		/* depth */
		" -1"		/* pen_style - not used */
		" -1"		/* area_fill: no fill */
		" 1.000"	/* style_val (1/80 inch) */
		" 0"		/* join_style: mitre */
		" 0"		/* cap_style: butt */
		" -1"		/* radius */
		" 0"		/* forward_arrow */
		" 0"		/* backward_arrow */
		" 2"		/* npoints */
		"\n");

	_debug("VDOTS: (%d,%d) -> (%d,%d)",
	       vdot->top, vdot->col, vdot->bottom, vdot->col);

	fprintf(out, "\t");
	fprintf(out, " %d %d", XSCALEH(vdot->col), YSCALE(vdot->top) - YSCALE(1) / 4);
	fprintf(out, " %d %d", XSCALEH(vdot->col), YSCALE(vdot->bottom + 1) + YSCALE(1) / 4);
	fprintf(out, "\n");
}
